"""Terms and Conditions for using the CIDC CLI."""

import click

from . import cache

# Copied verbatim from cli.interface.cli.CIDCCLI.info
TERMS = """
Welcome to the CIDC Command Line Interface (CLI) Tool.

You are about to access a system which contains data protected by federal law.

Unauthorized use of this system is strictly prohibited and subject to criminal
and civil penalties. All information stored on this system is owned by the
National Cancer Institute (NCI)."


By using this tool, you consent to the monitoring and recording of your
actions on this system. You also agree to refrain from engaging in any illegal
or improper behavior while using this system.

By downloading any data from the CIDC information system, you are agreeing to
take responsibility for the security of said data. You may not copy, transmit,
print out, or in any way cause the information to leave a secured computing
environment where it may be seen or accessed by unauthorized individuals.

Sharing your account with anyone else is strictly prohibited.

If you become aware of any threat to the system or possible breach of data,
you are required to immediately notify the CIDC.
"""

AGREEMENT = "Do you agree to the above terms and conditions?"

BOUNDARY = "=" * (len(AGREEMENT) + 10)

CONSENT_KEY = "consent"


def check_consent() -> bool:
    """Check that a user consents to the CLI user agreement."""
    # Has a user already consented?
    past_consent = cache.get(CONSENT_KEY)

    # If not, prompt the user to agree to the terms.
    if not past_consent:
        click.echo(TERMS)
        if click.confirm(AGREEMENT):
            # Store the user's response so we only need to ask them once.
            cache.store(CONSENT_KEY, "consented")
            click.echo("\nYour response has been stored.")
            click.echo(BOUNDARY)
        else:
            return False

    return True



"""The second generation CIDC command-line interface."""
import click

from . import api, auth, gcloud, upload, config, consent, __version__
from .dbedit.cli import get_username, list_, remove_, set_username

#### $ nci-cidc ####
@click.group(name="nci-cidc")
@click.option("-i", "--ignore", default=None, hidden=True)
def cidc(ignore):
    """The CIDC command-line interface."""
    config.check_env_warning(ignore)
    if not consent.check_consent():
        exit(0)
    gcloud.check_installed()


#### $ nci-cidc version ####
@click.command()
def version():
    """Echo version and done"""
    click.echo(f"nci-cidc-cli {__version__}")


#### $ nci-cidc login ####
@click.command()
@click.argument("portal_token", required=True, type=str)
def login(portal_token):
    """Validate and cache the given token"""
    click.echo("Validating token...")
    auth.validate_and_cache_token(portal_token)
    click.echo("You are now logged in.")


#### $ nci-cidc config ####
@click.group("config", hidden=True)
def config_():
    """Manage CLI configuration."""


#### $ nci-cidc config set-env ####
@click.command()
@click.argument(
    "environment",
    required=True,
    type=click.Choice(["prod", "staging", "dev-int", "dev", "dev-cloudtwo", "stage-cloudtwo", "prod-cloudtwo"]),
)
def set_env(environment):
    """Set the CLI environment."""
    config.set_env(environment)
    click.echo(f"Updated CLI environment to {environment}")


#### $ nci-cidc config get-env ####
@click.command()
def get_env():
    """Get the current CLI environment."""
    click.echo(config.get_env())


#### $ nci-cidc admin ####
@click.group("admin", hidden=True)
def admin_():
    """Manage API admin features."""


#### $ nci-cidc admin test-csms ####
@click.command()
def test_csms():
    """A simple API hit for a test of CSMS connection"""
    api.test_csms()


#### $ nci-cidc refresh-trial ####
@click.command()
@click.argument("trial_id", required=True, type=str)
def refresh_trial(trial_id):
    """refresh trial and associated files"""
    api.refresh_trial(trial_id)
    click.echo(f"Trial {trial_id} refreshed.")


#### $ nci-cidc assays ####
@click.group()
def assays():
    """Manage assay data."""


#### $ nci-cidc assays list ####
@click.command("list")
def list_assays():
    """List all supported assay types."""
    assay_list = api.list_assays()
    for assay in assay_list:
        click.echo(f"* {assay}")


#### $ nci-cidc assays upload ####
@click.command("upload")
@click.option("--assay", required=True, help="Assay type.")
@click.option("--xlsx", required=True, help="Path to the assay metadata spreadsheet.")
@click.option("--testing", is_flag=True, default=False, help="Flag for BDD testing.")
def upload_assay(assay, xlsx, testing):
    """
    Upload data for an assay.
    """
    upload.run_upload(assay, xlsx, is_testing=testing)


#### $ nci-cidc analyses ####
@click.group()
def analyses():
    """Manage analysis data."""


#### $ nci-cidc analyses list ####
@click.command("list")
def list_analyses():
    """List all supported analysis types."""
    analysis_list = api.list_analyses()
    for analysis in analysis_list:
        click.echo(f"* {analysis}")


#### $ nci-cidc analyses upload ####
@click.command("upload")
@click.option("--analysis", required=True, help="Analysis type.")
@click.option(
    "--xlsx", required=True, help="Path to the analysis metadata spreadsheet."
)
@click.option("--testing", is_flag=True, default=False, help="Flag for BDD testing.")
def upload_analysis(analysis, xlsx, testing):
    """
    Upload data for an analysis.
    """
    upload.run_upload(analysis, xlsx, is_analysis=True, is_testing=testing)


# Wire up the interface
cidc.add_command(version)
cidc.add_command(login)
cidc.add_command(assays)
cidc.add_command(analyses)
cidc.add_command(config_)
cidc.add_command(admin_)
cidc.add_command(refresh_trial)

config_.add_command(set_env)
config_.add_command(get_env)

assays.add_command(list_assays)
assays.add_command(upload_assay)

analyses.add_command(list_analyses)
analyses.add_command(upload_analysis)

admin_.add_command(test_csms)
admin_.add_command(get_username)
admin_.add_command(list_)
admin_.add_command(remove_)
admin_.add_command(set_username)

if __name__ == "__main__":
    cidc()



from typing import Tuple
import click

from . import core, remove, list as dblist
from . import config


#### $ nci-cidc admin get-username ####
@click.command()
def get_username():
    """Get the current database username."""
    click.echo(" ".join(["database username:", config.get_username()]))


#### $ nci-cidc admin set-username ####
@click.command()
@click.argument("username", required=True, type=str)
def set_username(username: str):
    """Set the database username."""
    config.set_username(username)
    click.echo(f"Updated database username to {username}")


#### $ nci-cidc admin list ####
@click.group("list")
def list_():
    """Listing different things"""
    pass


#### $ nci-cidc admin list supported ####
@click.command("supported")
def list_supported():
    """List assays and analyses that are supported for listing"""
    print(", ".join(sorted(list(dblist.SUPPORTED_ASSAYS_AND_ANALYSES))))


#### $ nci-cidc admin list assay ####
@click.command("assay")
@click.argument("trial_id", required=True, type=str)
@click.argument("assay_or_analysis", required=True, type=str)
def list_assay(trial_id: str, assay_or_analysis: str):
    """
    List CIMAC IDs for a given assay or analysis for a given trial
    Same as `nci-cidc admin list analysis`

    TRIAL_ID is the id of the trial to affect
    ASSAY_OR_ANALYSIS is the assay or analysis to list CIMAC IDs for
    """
    core.connect(dblist)
    dblist.list_data_cimac_ids(trial_id=trial_id, assay_or_analysis=assay_or_analysis)


#### $ nci-cidc admin list clinical ####
@click.command("clinical")
@click.argument("trial_id", required=True, type=str)
def list_clinical(trial_id: str):
    """
    List clinical files for a given trial

    TRIAL_ID is the id of the trial to affect
    """
    core.connect(dblist)
    dblist.list_clinical(trial_id=trial_id)


#### $ nci-cidc admin list misc-data ####
@click.command("misc-data")
@click.argument("trial_id", required=True, type=str)
def list_misc_data(trial_id: str):
    """
    List files from misc_data uploads for a given trial

    TRIAL_ID is the id of the trial to affect
    """
    core.connect(dblist)
    dblist.list_misc_data(trial_id=trial_id)


#### $ nci-cidc admin list shipments ####
@click.command("shipments")
@click.argument("trial_id", required=True, type=str)
def list_shipments(trial_id: str):
    """
    List shipments for a given trial

    TRIAL_ID is the id of the trial to affect
    """
    core.connect(dblist)
    dblist.list_shipments(trial_id=trial_id)


#### $ nci-cidc admin remove ####
@click.group("remove")
def remove_():
    """Deleting different things"""
    pass


#### $ nci-cidc admin remove assay ####
@click.command("assay")
@click.argument("trial_id", required=True, type=str)
@click.argument("assay_or_analysis", required=True, type=str)
@click.argument("target_id", required=True, nargs=-1)
def remove_assay(trial_id: str, assay_or_analysis: str, target_id: Tuple[str]):
    """
    Remove a given clinical data file from a given trial's metadata
    as well as remove the associated files themselves from the portal.

    TRIAL_ID is the id of the trial to affect
    ASSAY_OR_ANALYSIS is the assay or analysis to affect
    TARGET_ID is a tuple of the ids to find the data to remove
        it cannot go past where is divisible in the data
        eg if ASSAY_OR_ANALYSIS == "elisa", only `assay_run_id` is accepted
        eg if ASSAY_OR_ANALYSIS == "wes_analysis", only `run_id` is accepted
        eg if ASSAY_OR_ANALYSIS == "olink", `batch_id [file]` is assumed
    """
    core.connect(remove)
    remove.remove_data(
        trial_id=trial_id, assay_or_analysis=assay_or_analysis, target_id=target_id
    )


#### $ nci-cidc admin remove clinical ####
@click.command("clinical")
@click.argument("trial_id", required=True, type=str)
@click.argument("target_id", required=True, type=str)
def remove_clinical(trial_id: str, target_id: str):
    """
    Remove a given clinical data file from a given trial's metadata
    as well as remove the file itself from the portal.

    TRIAL_ID is the id of the trial to affect
    TARGET_ID is the object_url of the clinical data to remove
        not including {trial_id}/clinical/
        special value * for all files for this trial
    """
    core.connect(remove)
    remove.remove_clinical(trial_id=trial_id, target_id=target_id)


#### $ nci-cidc admin remove shipment ####
@click.command("shipment")
@click.argument("trial_id", required=True, type=str)
@click.argument("target_id", required=True, type=str)
def remove_shipment(trial_id: str, target_id: str):
    """
    Remove a given shipment from a given trial's metadata

    TRIAL_ID is the id of the trial to affect
    TARGET_ID is the manifest_id of the shipment to remove
    """
    core.connect(remove)
    remove.remove_shipment(trial_id=trial_id, target_id=target_id)


list_.add_command(list_assay)
list_.add_command(list_clinical)
list_.add_command(list_misc_data)
list_.add_command(list_shipments)
list_.add_command(list_supported)
remove_.add_command(remove_assay)
remove_.add_command(remove_clinical)
remove_.add_command(remove_shipment)



from datetime import datetime
import pandas as pd
from _pytest.monkeypatch import MonkeyPatch
from unittest.mock import MagicMock

from nci_cidc_cli.dbedit import list as dbedit_list
from .constants import (
    TEST_CLINICAL_URL_CSV,
    TEST_CLINICAL_URL_XLSX,
    TEST_METADATA_JSON,
    TEST_MISC_DATA_URL1,
    TEST_MISC_DATA_URL2,
    TEST_TRIAL_ID,
)


def test_list_clinical(monkeypatch):
    Session = MagicMock()
    session = MagicMock()
    begin = MagicMock()
    begin.__enter__.return_value = session
    Session.begin.return_value = begin

    mock_trial = MagicMock()
    mock_trial.metadata_json = TEST_METADATA_JSON
    get_trial_if_exists = MagicMock()
    get_trial_if_exists.return_value = mock_trial

    file1, file2 = MagicMock(), MagicMock()
    file1.object_url = f"{TEST_TRIAL_ID}/clinical/{TEST_CLINICAL_URL_XLSX}"
    file2.object_url = f"{TEST_TRIAL_ID}/clinical/{TEST_CLINICAL_URL_CSV}"
    file1._created = datetime.fromisoformat("2020-01-01T12:34:45")
    file2._created = datetime.fromisoformat("2020-02-02T12:34:45")

    get_clinical_downloadable_files = MagicMock()
    get_clinical_downloadable_files.return_value = [file1, file2]

    mock_print = MagicMock()

    monkeypatch.setattr(dbedit_list, "Session", Session)
    monkeypatch.setattr(dbedit_list, "get_trial_if_exists", get_trial_if_exists)
    monkeypatch.setattr(
        dbedit_list, "get_clinical_downloadable_files", get_clinical_downloadable_files
    )
    monkeypatch.setattr("builtins.print", mock_print)

    dbedit_list.list_clinical(TEST_TRIAL_ID)

    Session.begin.assert_called_once_with()
    begin.__enter__.assert_called_once()
    get_clinical_downloadable_files.assert_called_once_with(
        TEST_TRIAL_ID, session=session
    )
    begin.__exit__.assert_called_once()

    mock_print.assert_called_once()
    args, _ = mock_print.call_args
    assert len(args) == 1

    df = args[0]
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (2, 5)
    assert all(
        col in df.columns
        for col in ["created", "filename", "num_participants", "object_url", "comment"]
    )

    assert df.equals(
        pd.DataFrame(
            [
                [
                    file1.object_url,
                    TEST_CLINICAL_URL_XLSX,
                    5,
                    file1._created,
                    "comment",
                ],
                [file2.object_url, TEST_CLINICAL_URL_CSV, 3, file2._created, pd.NA],
            ],
            columns=[
                "object_url",
                "filename",
                "num_participants",
                "created",
                "comment",
            ],
        )
    )


def test_list_misc_data(monkeypatch):
    Session = MagicMock()
    session = MagicMock()
    begin = MagicMock()
    begin.__enter__.return_value = session
    Session.begin.return_value = begin

    mock_trial = MagicMock()
    mock_trial.metadata_json = TEST_METADATA_JSON
    get_trial_if_exists = MagicMock()
    get_trial_if_exists.return_value = mock_trial

    file1, file2 = MagicMock(), MagicMock()
    file1.object_url = f"{TEST_TRIAL_ID}/misc_data/{TEST_MISC_DATA_URL1}"
    file2.object_url = f"{TEST_TRIAL_ID}/misc_data/{TEST_MISC_DATA_URL2}"
    file1._created = datetime.fromisoformat("2020-01-01T12:34:45")
    file2._created = datetime.fromisoformat("2020-02-02T12:34:45")

    get_misc_data_files = MagicMock()
    get_misc_data_files.return_value = [file1, file2]

    mock_print = MagicMock()

    monkeypatch.setattr(dbedit_list, "Session", Session)
    monkeypatch.setattr(dbedit_list, "get_trial_if_exists", get_trial_if_exists)
    monkeypatch.setattr(dbedit_list, "get_misc_data_files", get_misc_data_files)
    monkeypatch.setattr("builtins.print", mock_print)

    dbedit_list.list_misc_data(TEST_TRIAL_ID)

    Session.begin.assert_called_once_with()
    begin.__enter__.assert_called_once()
    get_misc_data_files.assert_called_once_with(TEST_TRIAL_ID, session=session)
    begin.__exit__.assert_called_once()

    mock_print.assert_called_once()
    args, _ = mock_print.call_args
    assert len(args) == 1

    df = args[0]
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (2, 5)
    assert all(
        col in df.columns
        for col in ["batch_id", "created", "description", "filename", "object_url"]
    )

    assert df.equals(
        pd.DataFrame(
            [
                [
                    0,
                    file1.object_url,
                    TEST_MISC_DATA_URL1,
                    file1._created,
                    "description",
                ],
                [1, file2.object_url, TEST_MISC_DATA_URL2, file2._created, pd.NA],
            ],
            columns=["batch_id", "object_url", "filename", "created", "description"],
        )
    )


def test_list_shipments(monkeypatch):
    Session = MagicMock()
    session = MagicMock()
    begin = MagicMock()
    begin.__enter__.return_value = session
    Session.begin.return_value = begin

    upload1, upload2 = MagicMock(), MagicMock()
    upload1.upload_type = "pbmc"
    upload2.upload_type = "plasma"
    upload1._created = datetime.fromisoformat("2020-01-01T12:34:45")
    upload2._created = datetime.fromisoformat("2020-02-02T12:34:45")

    upload1.metadata_patch = {
        "participants": [
            {"samples": [None] * 3},
            {"samples": [None]},
        ],
        "shipments": [{"manifest_id": "test_upload"}],
    }
    upload2.metadata_patch = {
        "participants": [
            {"samples": [None] * 3},
            {"samples": [None] * 2},
        ],
        "shipments": [{"manifest_id": "test_upload2"}],
    }

    get_shipments = MagicMock()
    get_shipments.return_value = [upload1, upload2]

    mock_print = MagicMock()
    monkeypatch.setattr(dbedit_list, "Session", Session)
    monkeypatch.setattr(dbedit_list, "get_shipments", get_shipments)
    monkeypatch.setattr("builtins.print", mock_print)

    dbedit_list.list_shipments(TEST_TRIAL_ID)

    Session.begin.assert_called_once_with()
    begin.__enter__.assert_called_once()
    get_shipments.assert_called_once_with(TEST_TRIAL_ID, session=session)
    begin.__exit__.assert_called_once()

    mock_print.assert_called_once()
    args, _ = mock_print.call_args
    assert len(args) == 1

    df = args[0]
    assert isinstance(df, pd.DataFrame)
    assert df.shape == (2, 4)
    assert all(
        col in df.columns
        for col in ["created", "manifest_id", "num_samples", "upload_type"]
    )

    assert df.equals(
        pd.DataFrame(
            [
                [upload1.upload_type, "test_upload", 4, upload1._created],
                [upload2.upload_type, "test_upload2", 5, upload2._created],
            ],
            columns=["upload_type", "manifest_id", "num_samples", "created"],
        )
    )


class Test_list_data_cimac_ids:
    def setup(self):
        self.Session = MagicMock()
        self.session = MagicMock()
        self.begin = MagicMock()
        self.begin.__enter__.return_value = self.session
        self.Session.begin.return_value = self.begin

        self.mock_trial = MagicMock()
        self.mock_trial.metadata_json = TEST_METADATA_JSON

        self.get_trial_if_exists = MagicMock()
        self.get_trial_if_exists.return_value = self.mock_trial

        self.mock_list_clinical = MagicMock()
        self.mock_list_misc_data = MagicMock()

        self.real_print = print
        self.mock_print = MagicMock()

        self.monkeypatch = MonkeyPatch()
        self.monkeypatch.setattr(dbedit_list, "Session", self.Session)
        self.monkeypatch.setattr(
            dbedit_list, "get_trial_if_exists", self.get_trial_if_exists
        )
        self.monkeypatch.setattr(dbedit_list, "list_clinical", self.mock_list_clinical)
        self.monkeypatch.setattr(
            dbedit_list, "list_misc_data", self.mock_list_misc_data
        )
        self.monkeypatch.setattr("builtins.print", self.mock_print)

    def test_bail_outs(self):
        dbedit_list.list_data_cimac_ids(trial_id="foo", assay_or_analysis="bar")
        self.mock_print.assert_called_once_with(
            "Assay / analysis not supported:", "bar"
        )

        self.mock_print.reset_mock()
        dbedit_list.list_data_cimac_ids(
            trial_id="foo", assay_or_analysis="clinical_data"
        )
        self.mock_print.assert_not_called()
        self.mock_list_clinical.assert_called_once_with("foo")

        self.mock_print.reset_mock()
        dbedit_list.list_data_cimac_ids(trial_id="foo", assay_or_analysis="misc_data")
        self.mock_print.assert_not_called()
        self.mock_list_misc_data.assert_called_once_with("foo")

    def _get_and_assert_df(self) -> pd.DataFrame:
        self.mock_print.assert_called_once()
        args, _ = self.mock_print.call_args
        assert len(args) == 1 and isinstance(args[0], pd.DataFrame)
        return args[0]

    def test_olink(self):
        self.mock_print.reset_mock()
        dbedit_list.list_data_cimac_ids(trial_id="foo", assay_or_analysis="olink")
        df: pd.DataFrame = self._get_and_assert_df()

        assert df.equals(
            pd.DataFrame(
                [
                    {
                        "batch_id": "olink_batch",
                        "file": "combined",
                        "cimac_id": "CTTTPP101.00",
                    },
                    {
                        "batch_id": "olink_batch_2",
                        "file": f"{TEST_TRIAL_ID}/olink/batch_olink_batch_2/chip_0/assay_npx.xlsx",
                        "cimac_id": "CTTTPP201.00",
                    },
                    {
                        "batch_id": "olink_batch_2",
                        "file": f"{TEST_TRIAL_ID}/olink/batch_olink_batch_2/chip_1/assay_npx.xlsx",
                        "cimac_id": "CTTTPP102.00",
                    },
                ]
            )
        )

    def test_elisa(self):
        self.mock_print.reset_mock()
        dbedit_list.list_data_cimac_ids(trial_id="foo", assay_or_analysis="elisa")
        df: pd.DataFrame = self._get_and_assert_df()

        assert df.equals(
            pd.DataFrame(
                [
                    {"assay_run_id": "elisa_batch", "cimac_id": "CTTTPP101.00"},
                    {"assay_run_id": "elisa_batch", "cimac_id": "CTTTPP201.00"},
                    {"assay_run_id": "elisa_batch_2", "cimac_id": "CTTTPP102.00"},
                ]
            )
        )

    def test_nanostring(self):
        self.mock_print.reset_mock()
        dbedit_list.list_data_cimac_ids(trial_id="foo", assay_or_analysis="nanostring")
        df: pd.DataFrame = self._get_and_assert_df()

        assert df.equals(
            pd.DataFrame(
                [
                    {
                        "batch_id": "nanostring_batch",
                        "run_id": "nanostring_batch_run",
                        "cimac_id": "CTTTPP101.00",
                    },
                    {
                        "batch_id": "nanostring_batch",
                        "run_id": "nanostring_batch_run_2",
                        "cimac_id": "CTTTPP201.00",
                    },
                    {
                        "batch_id": "nanostring_batch_2",
                        "run_id": "nanostring_batch_2_run",
                        "cimac_id": "CTTTPP102.00",
                    },
                ]
            )
        )

    def test_rna_level1_analysis(self):
        self.mock_print.reset_mock()
        dbedit_list.list_data_cimac_ids(
            trial_id="foo", assay_or_analysis="rna_level1_analysis"
        )
        df: pd.DataFrame = self._get_and_assert_df()

        assert df.equals(
            pd.DataFrame(
                [
                    {"cimac_id": cimac_id}
                    for cimac_id in ["CTTTPP101.00", "CTTTPP201.00", "CTTTPP102.00"]
                ]
            )
        )

    def test_wes_analysis(self):
        self.mock_print.reset_mock()
        dbedit_list.list_data_cimac_ids(
            trial_id="foo", assay_or_analysis="wes_analysis"
        )
        df: pd.DataFrame = self._get_and_assert_df()

        assert df.equals(
            pd.DataFrame(
                [
                    {
                        "run_id": "CTTTPP101.00",
                        "tumor_cimac_id": "CTTTPP101.00",
                        "normal_cimac_id": "CTTTPP10N.00",
                    },
                    {
                        "run_id": "CTTTPP102.00",
                        "tumor_cimac_id": "CTTTPP102.00",
                        "normal_cimac_id": "CTTTPP10N.00",
                    },
                    {
                        "run_id": "CTTTPP201.00",
                        "tumor_cimac_id": "CTTTPP201.00",
                        "normal_cimac_id": "CTTTPP20N.00",
                    },
                    {
                        "run_id": "CTTTPP201.00",
                        "tumor_cimac_id": "CTTTPP201.00",
                        "normal_cimac_id": "CTTTPP20N.00",
                    },
                    {
                        "run_id": "CTTTPP102.00",
                        "tumor_cimac_id": "CTTTPP102.00",
                        "normal_cimac_id": "CTTTPP10N.00",
                    },
                ]
            )
        )

    def test_wes_analysis_old(self):
        self.mock_print.reset_mock()
        dbedit_list.list_data_cimac_ids(
            trial_id="foo", assay_or_analysis="wes_analysis_old"
        )
        df: pd.DataFrame = self._get_and_assert_df()

        assert df.equals(
            pd.DataFrame(
                [
                    {
                        "run_id": "CTTTPP201.00",
                        "tumor_cimac_id": "CTTTPP201.00",
                        "normal_cimac_id": "CTTTPP20N.00",
                    },
                    {
                        "run_id": "CTTTPP102.00",
                        "tumor_cimac_id": "CTTTPP102.00",
                        "normal_cimac_id": "CTTTPP10N.00",
                    },
                ]
            )
        )

    def test_wes_tumor_only_analysis(self):
        self.mock_print.reset_mock()
        dbedit_list.list_data_cimac_ids(
            trial_id="foo", assay_or_analysis="wes_tumor_only_analysis"
        )
        df: pd.DataFrame = self._get_and_assert_df()

        assert df.equals(
            pd.DataFrame(
                [
                    {"cimac_id": cimac_id}
                    for cimac_id in [
                        "CTTTPP101.00",
                        "CTTTPP102.00",
                        "CTTTPP201.00",
                        "CTTTPP201.00",
                        "CTTTPP102.00",
                    ]
                ]
            )
        )

    def test_wes_tumor_only_analysis_old(self):
        self.mock_print.reset_mock()
        dbedit_list.list_data_cimac_ids(
            trial_id="foo", assay_or_analysis="wes_tumor_only_analysis_old"
        )
        df: pd.DataFrame = self._get_and_assert_df()

        assert df.equals(
            pd.DataFrame([{"cimac_id": "CTTTPP201.00"}, {"cimac_id": "CTTTPP102.00"}])
        )

    def test_cytof_analysis(self):
        self.mock_print.reset_mock()
        dbedit_list.list_data_cimac_ids(
            trial_id="foo", assay_or_analysis="cytof_analysis"
        )
        df: pd.DataFrame = self._get_and_assert_df()

        assert df.equals(
            pd.DataFrame(
                [
                    {
                        "batch_id": "cytof_batch",
                        "cimac_id": "CTTTPP101.00",
                    },
                    {
                        "batch_id": "cytof_batch",
                        "cimac_id": "CTTTPP201.00",
                    },
                    {
                        "batch_id": "cytof_batch_2",
                        "cimac_id": "CTTTPP102.00",
                    },
                ]
            )
        )

    def test_batched_analysis(self):
        self.mock_print.reset_mock()
        dbedit_list.list_data_cimac_ids(
            trial_id="foo", assay_or_analysis="tcr_analysis"
        )
        df: pd.DataFrame = self._get_and_assert_df()

        assert df.equals(
            pd.DataFrame(
                [
                    {
                        "batch_id": "tcr_analysis_batch",
                        "cimac_id": "CTTTPP101.00",
                    },
                    {
                        "batch_id": "tcr_analysis_batch",
                        "cimac_id": "CTTTPP201.00",
                    },
                    {
                        "batch_id": "tcr_analysis_batch_2",
                        "cimac_id": "CTTTPP102.00",
                    },
                ]
            )
        )

    def test_generic_assay_with_ids(self):
        self.mock_print.reset_mock()
        self.mock_print.reset_mock()
        dbedit_list.list_data_cimac_ids(
            trial_id="foo", assay_or_analysis="atacseq_analysis"
        )

        df: pd.Series = self._get_and_assert_df()

        assert df.equals(
            pd.DataFrame(
                [
                    {
                        "batch_id": "atacseq_analysis_batch",
                        "cimac_id": "CTTTPP101.00",
                    },
                    {
                        "batch_id": "atacseq_analysis_batch",
                        "cimac_id": "CTTTPP201.00",
                    },
                    {
                        "batch_id": "atacseq_analysis_batch_2",
                        "cimac_id": "CTTTPP102.00",
                    },
                ]
            )
        )

    def test_generic_assay_no_ids(self):
        self.mock_print.reset_mock()
        self.mock_print.reset_mock()
        dbedit_list.list_data_cimac_ids(trial_id="foo", assay_or_analysis="wes")

        df: pd.Series = self._get_and_assert_df()
        assert df.equals(
            pd.DataFrame(
                [
                    {
                        "batch_id": "0",
                        "cimac_id": "CTTTPP101.00",
                    },
                    {
                        "batch_id": "0",
                        "cimac_id": "CTTTPP201.00",
                    },
                    {
                        "batch_id": "1",
                        "cimac_id": "CTTTPP102.00",
                    },
                ]
            )
        )



from click.testing import CliRunner

from nci_cidc_cli import cli, consent, config, __version__
from functools import wraps


def skip_consent(monkeypatch):
    monkeypatch.setattr("nci_cidc_cli.consent.check_consent", lambda: True)


def skip_gcloud_installation_check(monkeypatch):
    monkeypatch.setattr("nci_cidc_cli.gcloud.check_installed", lambda: True)


def with_default_env(fn):
    """Ensures that a test run icurrent CLI config isn't overwritten by a test run"""

    @wraps(fn)
    def wrap(*args, **kwds):
        prev_env = config.get_env()
        config.set_env("prod")
        try:
            return fn(*args, **kwds)
        finally:
            config.set_env(prev_env)

    return wrap


@with_default_env
def test_cidc_structure(runner: CliRunner, monkeypatch):
    """
    Check that the interface is wired up correctly, and that
    usage is printed when commands are supplied without arguments.
    """
    skip_consent(monkeypatch)
    skip_gcloud_installation_check(monkeypatch)

    res = runner.invoke(cli.cidc)
    assert "Usage: nci-cidc" in res.output

    res = runner.invoke(cli.cidc, ["version"])
    assert f"nci-cidc-cli {__version__}" in res.output

    res = runner.invoke(cli.cidc, ["assays"])
    assert "Usage: nci-cidc assays" in res.output

    res = runner.invoke(cli.cidc, ["config"])
    assert "Usage: nci-cidc config" in res.output

    res = runner.invoke(cli.cidc, ["login", "-h"])
    assert "Usage: nci-cidc login" in res.output


@with_default_env
def test_no_gcloud_installation(runner: CliRunner, monkeypatch):
    """
    Check that running `nci-cidc [subcommand]` without a gcloud installation prompts
    the user to install gcloud.
    """
    skip_consent(monkeypatch)

    def assert_gcloud_message(res):
        assert "requires an installation of the gcloud SDK" in res.output

    monkeypatch.setattr("shutil.which", lambda *args: False)
    assert_gcloud_message(runner.invoke(cli.cidc, ["assays"]))
    assert_gcloud_message(runner.invoke(cli.cidc, ["login"]))


@with_default_env
def test_assays_list(runner: CliRunner, monkeypatch):
    """
    Check that assay_list displays supported assays as expected.
    """
    monkeypatch.setattr("nci_cidc_cli.api.list_assays", lambda: ["wes", "pbmc"])
    res = runner.invoke(cli.assays, ["list"])
    assert "* wes" in res.output
    assert "* pbmc" in res.output


def test_env_config(runner: CliRunner, monkeypatch):
    """
    Test setting and getting the current environment.
    """
    monkeypatch.setattr("nci_cidc_cli.cache._cache_dir", lambda: "workdir")
    with runner.isolated_filesystem():
        # Get default value
        res = runner.invoke(cli.get_env)
        assert "prod" in res.output

        # Set to valid value
        res = runner.invoke(cli.set_env, ["dev"])
        assert "dev" in res.output
        res = runner.invoke(cli.get_env)
        assert "dev" in res.output

        # Try to set to invalid value
        res = runner.invoke(cli.set_env, ["blah"])
        assert "Invalid value" in res.output


def test_refresh_trial(runner: CliRunner, monkeypatch):
    """
    Test refreshing a trial command
    """
    monkeypatch.setattr("nci_cidc_cli.api.refresh_trial", lambda x: "Response")
    trial_id = "SAMPLE_TRIAL_ID"
    res = runner.invoke(cli.refresh_trial, [trial_id])
    assert trial_id in res.output


def test_consent(runner: CliRunner, monkeypatch):
    """Check the consent flow."""
    monkeypatch.setattr("nci_cidc_cli.cache._cache_dir", lambda: "workdir")
    with runner.isolated_filesystem():
        # User has not yet consented, so prompt for consent.
        res = runner.invoke(cli.cidc, ["assays"], input="y")
        assert consent.TERMS in res.output
        assert consent.AGREEMENT in res.output

        # User has now consented, so don't prompt.
        res = runner.invoke(cli.cidc, ["assays"])
        assert consent.TERMS not in res.output


def test_assays_upload(runner: CliRunner, monkeypatch):
    pass



"""Methods for working with id tokens"""
import click
from jose import jwt
from jose.exceptions import JWTError

from . import api
from . import cache


TOKEN = "id_token"


class AuthError(click.ClickException):
    pass


def unauthenticated() -> AuthError:
    return AuthError(
        "You are not authenticated. Please login with:\n"
        "   $ nci-cidc login [YOUR PORTAL TOKEN]"
    )


def validate_token(id_token: str):
    """
    Raises AuthError if id_token is not valid
    """
    try:
        error = api.check_auth(id_token)
    except api.ApiError as e:
        raise AuthError(str(e))


def validate_and_cache_token(id_token: str):
    """
    If a token is valid, cache it for use in future commands.
    """
    # Validate the id token
    validate_token(id_token)

    # Save the provided token
    cache.store(TOKEN, id_token)


def get_id_token() -> str:
    """
    Look for a cached id_token for this user. If no token is cached,
    exit and prompt the user to log in. Otherwise, return the cached token.
    """
    # Try to find a cached token
    id_token = cache.get(TOKEN)

    # If there's no cached token, the user needs to log in
    if not id_token:
        raise unauthenticated()

    return id_token


def get_user_email() -> str:
    """Extract a user's email from their id token."""
    token = get_id_token()

    # We don't need to check verifications here,
    # because get_id_token validates the token it returns
    # with the API (this includes signature verification).
    try:
        claims = jwt.get_unverified_claims(token)
    except JWTError:
        raise unauthenticated()

    return claims["email"]



import pytest
from unittest.mock import MagicMock

from nci_cidc_cli.dbedit import core
from nci_cidc_cli.config import set_env

from .constants import TEST_PASSWORD, TEST_TRIAL_ID, TEST_USER


class Mocker:
    def __init__(self, monkeypatch) -> None:
        self.sqlalchemy = MagicMock()

        self.Connector = MagicMock()
        self.Connector_instance = MagicMock()
        self.Connector.return_value = self.Connector_instance

        self.sessionmaker = MagicMock()
        self.Session = MagicMock()
        self.sessionmaker.return_value = self.Session

        self.automap_base = MagicMock()
        self.Base = MagicMock()

        self.DownloadableFiles = MagicMock()
        self.TrialMetadata = MagicMock()
        self.UploadJobs = MagicMock()
        self.Users = MagicMock()

        self.Base.classes.downloadable_files = self.DownloadableFiles
        self.Base.classes.trial_metadata = self.TrialMetadata
        self.Base.classes.upload_jobs = self.UploadJobs
        self.Base.classes.users = self.Users
        self.automap_base.return_value = self.Base

        self.create_engine_result = MagicMock()

        def create_engine(*args, creator: callable, **kwargs) -> MagicMock:
            creator()
            return self.create_engine_result

        self.sqlalchemy.create_engine = MagicMock(wraps=create_engine)

        monkeypatch.setattr(core, "Connector", self.Connector)
        monkeypatch.setattr(core, "sqlalchemy", self.sqlalchemy)
        monkeypatch.setattr(core, "automap_base", self.automap_base)
        monkeypatch.setattr(core, "sessionmaker", self.sessionmaker)

        monkeypatch.setattr("builtins.input", lambda _: TEST_USER)
        monkeypatch.setattr(core.getpass, "getpass", lambda: TEST_PASSWORD)

    def reset_mocks(self):
        self.Connector.reset_mock()
        self.sqlalchemy.reset_mock()
        self.sqlalchemy.create_engine.reset_mock()
        self.automap_base.reset_mock()
        self.sessionmaker.reset_mock()


def test_connect(monkeypatch):
    monkeypatch.setattr(core, "get_env", lambda: "prod")
    monkeypatch.setattr(core, "get_username", lambda: TEST_USER)
    mocks = Mocker(monkeypatch)

    assert core.DownloadableFiles is None
    assert core.Session is None
    assert core.TrialMetadata is None
    assert core.UploadJobs is None
    assert core.Users is None

    list_mod = MagicMock()
    core.connect(list_mod)

    mocks.Connector.assert_called_once_with()
    mocks.automap_base.assert_called_once_with()
    mocks.sqlalchemy.create_engine.assert_called_once()
    args, _ = mocks.sqlalchemy.create_engine.call_args
    assert args == ("postgresql+pg8000://",)

    mocks.Connector_instance.connect.assert_called_once_with(
        "nih-nci-cidc-datapipelines:us-east4:cidc-postgresql-cidc-prod2",
        "pg8000",
        user=TEST_USER,
        password=TEST_PASSWORD,
        db="cidc-cidc-prod2",
    )

    mocks.sessionmaker.assert_called_once_with(mocks.create_engine_result)
    mocks.Base.prepare.assert_called_once()
    args, _ = mocks.Base.prepare.call_args
    assert args == (mocks.create_engine_result,)

    assert core.DownloadableFiles is not None
    assert core.Session is not None
    assert core.TrialMetadata is not None
    assert core.UploadJobs is not None
    assert core.Users is not None

    assert list_mod.DownloadableFiles is not None
    assert list_mod.Session is not None
    assert list_mod.TrialMetadata is not None
    assert list_mod.UploadJobs is not None
    assert list_mod.Users is not None

    mocks.reset_mocks()
    list_mod.reset_mock()
    # check that it switches to dev-int if no ENV
    monkeypatch.setattr(core, "get_env", lambda: None)
    core.connect(list_mod)
    mocks.Connector_instance.connect.assert_called_once_with(
        "nih-nci-cimac-cidc-dpline-dev:us-east4:cidc-postgresql-cidc-dev2",
        "pg8000",
        user=TEST_USER,
        password=TEST_PASSWORD,
        db="cidc-cidc-dev2",
    )
    mocks.reset_mocks()
    list_mod.reset_mock()
    # check that it switches to dev-int if weird ENV
    monkeypatch.setattr(core, "get_env", lambda: "foo")
    core.connect(list_mod)
    mocks.Connector_instance.connect.assert_called_once_with(
        "nih-nci-cimac-cidc-dpline-dev:us-east4:cidc-postgresql-cidc-dev2",
        "pg8000",
        user=TEST_USER,
        password=TEST_PASSWORD,
        db="cidc-cidc-dev2",
    )


def test_get_clinical_downloadable_files(monkeypatch):
    monkeypatch.setattr(core, "get_env", lambda: "dev")
    DownloadableFiles = MagicMock()
    monkeypatch.setattr(core, "DownloadableFiles", DownloadableFiles)

    session = MagicMock()
    # these are to hold the results
    query = MagicMock()
    query_filter = MagicMock()

    mock_files = [MagicMock(), MagicMock()]
    query_filter.all.return_value = mock_files

    query.filter.return_value = query_filter
    session.query.return_value = query

    res: list = core.get_clinical_downloadable_files(TEST_TRIAL_ID, session=session)
    assert len(res) == 2 and res is mock_files

    session.query.assert_called_once_with(DownloadableFiles)
    query.filter.assert_called_once_with(
        DownloadableFiles.trial_id == TEST_TRIAL_ID,
        DownloadableFiles.object_url.like("%/clinical/%"),
    )
    query_filter.all.assert_called_once_with()


def test_get_misc_data_files(monkeypatch):
    monkeypatch.setattr(core, "get_env", lambda: "dev")
    DownloadableFiles = MagicMock()
    monkeypatch.setattr(core, "DownloadableFiles", DownloadableFiles)

    session = MagicMock()
    # these are to hold the results
    query = MagicMock()
    query_filter = MagicMock()

    mock_files = [MagicMock(), MagicMock()]
    query_filter.all.return_value = mock_files

    query.filter.return_value = query_filter
    session.query.return_value = query

    res: list = core.get_misc_data_files(TEST_TRIAL_ID, session=session)
    assert len(res) == 2 and res is mock_files

    session.query.assert_called_once_with(DownloadableFiles)
    query.filter.assert_called_once_with(
        DownloadableFiles.trial_id == TEST_TRIAL_ID,
        DownloadableFiles.object_url.like("%/misc_data/%"),
    )
    query_filter.all.assert_called_once_with()


def test_get_shipments(monkeypatch):
    monkeypatch.setattr(core, "get_env", lambda: "dev")
    UploadJobs = MagicMock()
    monkeypatch.setattr(core, "UploadJobs", UploadJobs)

    session = MagicMock()
    # these are to hold the results
    query = MagicMock()
    query_filter = MagicMock()

    mock_uploads = [MagicMock(), MagicMock()]
    mock_uploads[0].metadata_patch = {"shipments": []}  # should be returned
    mock_uploads[1].metadata_patch = {}  # should NOT be returned
    query_filter.all.return_value = mock_uploads

    query.filter.return_value = query_filter
    session.query.return_value = query

    res: list = core.get_shipments(TEST_TRIAL_ID, session=session)
    assert len(res) == 1 and res[0] is mock_uploads[0]

    session.query.assert_called_once_with(UploadJobs)
    query.filter.assert_called_once_with(
        UploadJobs.trial_id == TEST_TRIAL_ID,
        UploadJobs.status == "merge-completed",
    )
    query_filter.all.assert_called_once_with()


def test_get_trial_if_exists(monkeypatch):
    monkeypatch.setattr(core, "get_env", lambda: "dev")
    Session = MagicMock()
    session = MagicMock()
    begin = MagicMock()

    TrialMetadata = MagicMock()
    monkeypatch.setattr(core, "TrialMetadata", TrialMetadata)

    # these are to hold the results
    query = MagicMock()
    query_filter = MagicMock()
    with_for_update = MagicMock()

    mock_trial = MagicMock()
    mock_trial.metadata_json = {"foo": "bar"}

    with_for_update.first.return_value = mock_trial
    query_filter.with_for_update.return_value = with_for_update
    query.filter.return_value = query_filter
    session.query.return_value = query
    begin.__enter__.return_value = session
    Session.begin.return_value = begin
    monkeypatch.setattr(core, "Session", Session)

    def reset_mocks():
        Session.reset_mock()
        session.reset_mock()
        begin.reset_mock()
        TrialMetadata.reset_mock()
        query.reset_mock()
        query_filter.reset_mock()
        with_for_update.reset_mock()
        mock_trial.reset_mock()

    # check it returns the trial for update
    with_for_update.first.return_value = mock_trial
    query_filter.with_for_update.return_value = with_for_update
    query.filter.return_value = query_filter
    session.query.return_value = query
    begin.__enter__.return_value = session
    Session.begin.return_value = begin
    monkeypatch.setattr(core, "Session", Session)

    reset_mocks()
    trial: TrialMetadata = core.get_trial_if_exists(
        trial_id=TEST_TRIAL_ID,
        with_for_update=True,
        session=session,
    )
    assert trial is mock_trial

    # check that it exits if trial not found for update
    with_for_update.first.return_value = None
    query_filter.with_for_update.return_value = with_for_update
    query.filter.return_value = query_filter
    session.query.return_value = query
    begin.__enter__.return_value = session
    Session.begin.return_value = begin
    monkeypatch.setattr(core, "Session", Session)

    reset_mocks()
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        core.get_trial_if_exists(
            trial_id=TEST_TRIAL_ID,
            with_for_update=True,
            session=session,
        )
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0

    # check it returns the trial NOT for update
    query_filter.first.return_value = mock_trial
    query.filter.return_value = query_filter
    session.query.return_value = query
    begin.__enter__.return_value = session
    Session.begin.return_value = begin
    monkeypatch.setattr(core, "Session", Session)

    reset_mocks()
    trial: TrialMetadata = core.get_trial_if_exists(
        trial_id=TEST_TRIAL_ID,
        session=session,
    )
    assert trial is mock_trial

    # check that it exits if trial not found NOT for update
    query_filter.first.return_value = None
    query.filter.return_value = query_filter
    session.query.return_value = query
    begin.__enter__.return_value = session
    Session.begin.return_value = begin
    monkeypatch.setattr(core, "Session", Session)

    reset_mocks()
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        core.get_trial_if_exists(
            trial_id=TEST_TRIAL_ID,
            session=session,
        )
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0
